--      Copyright  2010.       GSI Technology
--                                              Jeff Daugherty
--                                              apps@gsitechnology.com
--  Version: 1.5
--
--      FileName: core_CIO_Plus.v
--
-- Revision History
-- 06/04/08    1.1    1) Updated model to removed GSI library.
-- 07/21/09    1.2    1) Model now has a reduced memory core.  The size is controlled by A_BITS
-- 09/29/09    1.3    1) Combined 2.0 and 2.5 latency models.
-- 11/15/10    1.4    1) Added capability to run reduced or normal array size
--                    2) Added clock init to prevent negative time assignments.
-- 12/15/10    1.5    1) Model supports both reduced and actual density memory core.  The size is controlled by A_BITS
--
LIBRARY ieee;
USE ieee.std_logic_1164.all;
ENTITY PLUS_CIO_B2_CORE IS
  GENERIC (
    CONSTANT A_size    : integer;
    constant mem_size  : integer;
    CONSTANT ECC         : integer := 0;
    constant A_bits    : integer;
    CONSTANT DQ_width  : integer;
    CONSTANT RL        : integer;
    constant MEM_STYLE : string(7 downto 1) := "NORMAL ");  -- REDUCED or NORMAL
  PORT (
    SIGNAL A         : in    std_logic_vector(A_size-1 DOWNTO 0);  -- address
    SIGNAL DQ        : inout std_logic_vector(DQ_width - 1 DOWNTO 0);  -- byte A data
    SIGNAL cBW       : in    std_logic_vector(3 downto 0);   -- bank 1 write enable
    SIGNAL K         : in    std_logic;   -- input clock
    SIGNAL nK        : in    std_logic;   -- input clock bar
    SIGNAL nLD       : in    std_logic;   -- Read enable
    SIGNAL RnW       : in    std_logic;   -- Write strobe
    SIGNAL nDoff      : in    std_logic;   -- Write strobe
    SIGNAL TCK       : in    std_logic;   -- Write strobe
    SIGNAL CQ        : out   std_logic;   -- Write strobe
    SIGNAL nCQ       : out   std_logic;   -- Write strobe
    SIGNAL QVLD      : out   std_logic;   -- Write strobe
    SIGNAL HighZ     :       std_logic_vector(DQ_width - 1 downto 0);
    SIGNAL tCHQV     :       time;
    SIGNAL tCHQX     :       time;
    SIGNAL tCQHQV    :       time;
    SIGNAL tCQHQX    :       time;
    SIGNAL tCHQV_off :       time;
    SIGNAL tCHQX_off :       time);
END PLUS_CIO_B2_CORE;

LIBRARY ieee;
ARCHITECTURE PLUS_CIO_B2 OF PLUS_CIO_B2_CORE IS
  use ieee.std_logic_unsigned.all, ieee.std_logic_arith.all;

  function pick (DQ_width: integer) return integer is
  begin
    if DQ_width = 18 or DQ_width = 8 then
      return 2;
    elsif DQ_width = 36 then
      return 4;
    else
      return 1;
    end if;
  end pick;

  type GSI_DATA_MEMORY is protected
    procedure set (variable address : integer;
                            data    : std_logic_vector(DQ_width - 1 downto 0);
                            bwi     : std_logic_vector(3 downto 0));
    procedure get (variable address : integer;
                   variable data    : out std_logic_vector(DQ_width - 1 downto 0));
  end protected GSI_DATA_MEMORY;

  type GSI_DATA_MEMORY is protected body
    type MEM is array (0 to 2**mem_size - 1) of std_logic_vector(DQ_width - 1 downto 0);
    variable MEMORY : MEM;
    procedure set (variable address : integer;
                            data    : std_logic_vector(DQ_width - 1 downto 0);
                            bwi     : std_logic_vector(3 downto 0)
                   ) is
      variable bw_size : integer := pick(DQ_width);
    begin
      for store in 0 to BW_size - 1 loop
        if bwi(store) = '1' then
          if DQ_width = 8 then
            MEMORY(address)(((store+1)*(DQ_width/2)) - 1 downto store*(DQ_width/2)) := data(((store+1)*(DQ_width/2)) - 1 downto store*(DQ_width/2));
          else
            MEMORY(address)(((store+1)*9) - 1 downto store*9) := data(((store+1)*9) - 1 downto store*9);
          end if;
        end if;
      end loop;
    end procedure set;
    procedure get(variable address : integer;
                  variable data : out std_logic_vector(DQ_width - 1 downto 0)
                  ) is
      begin
         data := MEMORY(address);
    end procedure get;
  end protected body GSI_DATA_MEMORY;
  
  type GSI_ADDR_MEMORY is protected
    procedure set (address : integer; variable data :     std_logic_vector(A_size downto 0));
    procedure get (address : integer; variable data : out std_logic_vector(A_size downto 0));
  end protected GSI_ADDR_MEMORY;

  type GSI_ADDR_MEMORY is protected body
    type MEM is array (0 to 2**A_bits) of std_logic_vector(A_size downto 0);
    variable MEMORY : MEM;
    procedure set (         address : integer;
                   variable data    : std_logic_vector(A_size downto 0)
                  ) is
      begin
        MEMORY(address) := data;
    end procedure set;
    procedure get(address : integer;
                  variable data : out std_logic_vector(A_size downto 0)
                  ) is
      begin
         data := MEMORY(address);
    end procedure get;
  end protected body GSI_ADDR_MEMORY;

  TYPE iDATA   is array (0 to 7) of std_logic_vector(DQ_width - 1 DOWNTO 0);
  TYPE ADDRESS is array (0 to 7) of std_logic_vector(A_size DOWNTO 0);
  TYPE BYTE    is array (0 to 7) of std_logic_vector(3 DOWNTO 0);

  shared variable MEMORY_CORE    : GSI_DATA_MEMORY;
  shared variable ADDRESS_CORE   : GSI_ADDR_MEMORY;

  signal Di            : iDATA;
  signal bw_i          : BYTE;  -- internal byte write enable

  constant reset       : std_logic_vector(A_size downto 0) := conv_std_logic_vector(0, A_size+1);
  SIGNAL Qi            : std_logic_vector(DQ_width-1 DOWNTO 0) ;
  SIGNAL Qi_o          : std_logic_vector(DQ_width-1 DOWNTO 0) ;
  SIGNAL rA            : std_logic_vector(A_size DOWNTO 0) ;
  SIGNAL read_pos      : std_logic_vector(7 DOWNTO 0) ;
  signal bwerr_flag    : std_logic_vector(3 downto 0) :="0000";
  signal we            : std_logic_vector(3 downto 0);

  SIGNAL W             : std_logic := '0';
  SIGNAL R             : std_logic := '0';
  SIGNAL re            : std_logic := '0';
  SIGNAL R1            : std_logic := '0';
  SIGNAL K_i           : std_logic := '0';
  SIGNAL Q_switch      : std_logic := '0';
  SIGNAL CQ_o          : std_logic := 'Z';
  SIGNAL nCQ_o         : std_logic := 'Z';
  SIGNAL DLL           : std_logic;
  SIGNAL K_Time        : time    := 1 ns;
  SIGNAL nK_Time       : time    := 1 ns;
  SIGNAL CHQV          : time    := 1 ns;
  SIGNAL CHQX          : time    := 1 ns;
  SIGNAL CHCQon        : time    := 1 ns;
  SIGNAL CHCQoff       : time    := 1 ns;
  SIGNAL Delay         : integer := 0;
  SIGNAL Latency       : integer := 0;
  signal TCK_Count     : integer :=0;
  signal JTAG_Flag     : integer :=0;
  SIGNAL Read_Command  : STRING (4 downto 1);
  SIGNAL Write_Command : STRING (5 downto 1);

  signal wA_D  : ADDRESS := (reset, reset, reset, reset, reset, reset, reset, reset);
  signal rA_D  : ADDRESS := (reset, reset, reset, reset, reset, reset, reset, reset);

-- ---------------------------------------------------------------
--   	reduced memory stuff
-- ---------------------------------------------------------------
  signal address_used : std_logic_vector(a_bits-1 downto 0) := (others=>'0');
  constant MAX_BITS   : std_logic_vector(a_bits-1 downto 0) := (others=>'1');

--  TIMING FUNCTIONS
  function POSEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '0') OR (s = '1')));
  end; 
  function NEGEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '1') OR (s = '0')) );
  end; 
-- END TIMING FUNCTIONS

BEGIN
  -- purpose: Main K controlled process
  -- type   : combinational
  -- inputs : K, nK, Di, bw_i, DLL, A, 
  -- outputs: Qi, re, CQ_i, nCQ_i
  Main : PROCESS(K_i)
    -- purpose: shift address arrays
    -- inputs : wA_D, rA_D, read_pos
    -- outputs: wA_D, rA_D, read_pos
    procedure shift is
      variable Temp_write : Address;
      variable Temp_read  : Address;
      variable Temp_data  : iDATA;
      variable Temp_byte  : BYTE;
      variable i : integer;
    begin  -- shift
      Temp_write   := wA_D;
      Temp_read    := rA_D;
      Temp_data    := Di;
      Temp_byte    := BW_i;
      for i in 7 downto 1 loop
        Temp_write(i) := Temp_write(i-1);
        Temp_read(i)  := Temp_read(i-1);
        Temp_data(i)  := Temp_data(i-1);
        Temp_byte(i)  := Temp_byte(i-1);
      end loop;  -- i
      wA_D <= Temp_write;
      rA_D <= Temp_read;
      Di   <= Temp_data;
      bw_i <= Temp_byte;
    end procedure shift;

    -- purpose: Determine Read / Write / NOOP State
    -- inputs : E, W, RnW, R, R1, A
    -- outputs: Read_Command, Write_Command, Read_Pos, wA_D, rA_D, we, W, R, R1
    procedure State is
    begin
      if (RnW='1' and nLD='0') then
        Read_Command <= "READ";
        re <= '1';
        rA_D(1) <= (A(A_size-1 downto 0) & "0");
        rA_D(0) <= (A(A_size-1 downto 0) & "1");
      else
        Read_Command <= "NOOP";
        re <= '0';
      end if;
      if (RnW='0' and R='0' and R1='0' and nLD='0') then
        we(0) <= '1';
        Write_Command <= "Write";
        wA_D(1) <= (A(A_size-1 downto 0) & "0");
        wA_D(0) <= (A(A_size-1 downto 0) & "1");
      else
        we(0) <= '0';
        Write_Command <= "NO-OP";
      end if;
      W  <= not RnW;
      R1 <= R;
      R  <= RnW;
    end procedure State;

-- purpose: Read from array and set Qi
-- inputs : read_pos, re
-- outputs: re, Q_switch, Qi
    procedure Read_Array is
      variable index          : integer := 0;
      variable address_match  : integer := 0;
      variable address_index  : integer := 0;
      variable address_fetch  : std_logic_vector(A_size downto 0);
      variable data_fetch     : std_logic_vector(DQ_width - 1 downto 0);
    begin
      if MEM_STYLE = "REDUCED" then
        if conv_integer(address_used) > 0 then
          for index in 0 to conv_integer(address_used) - 1 loop
            ADDRESS_CORE.get(index, address_fetch);
            if (address_fetch = rA_D(latency) and address_match = 0) then
              address_match := 1;
              address_index := index;
            end if;
          end loop;  -- indev
        end if;
        if address_match = 1 then
          address_match := 0;
          MEMORY_CORE.get(address_index, data_fetch);
        else
          data_fetch := (others=>'X');
        end if;
      else
        address_index := conv_integer(rA_D(latency));
        MEMORY_CORE.get(address_index, data_fetch);
      end if;
      Qi <= data_fetch;
    end procedure Read_Array;

-- purpose: Write data to Array
-- inputs : bw_i, W
-- outputs: 
    procedure Write_Array(FLAG : STRING(4 downto 1)) is
      variable addr_t         : integer;
      variable index          : integer := 0;
      variable address_match  : integer := 0;
      variable address_index  : integer := 0;
      variable address_fetch  : std_logic_vector(A_size downto 0);
      variable address_store  : std_logic_vector(A_size downto 0);
    begin
      we(3) <= we(2);
      we(2) <= we(1);
      we(1) <= we(0);
      address_store := wA_D(4);
      if (we(3)='1')  then
        if MEM_STYLE = "REDUCED" then
          if conv_integer(address_used) >0 then
            for index in 0 to conv_integer(address_used)-1 loop
              ADDRESS_CORE.get(index, address_fetch);
              if (address_fetch = wA_D(4) and address_match=0) then
                address_match := 1;
                address_index := index;
              end if;
            end loop;  -- index
          end if;
          if address_match = 1 then
            address_match := 0;
            ADDRESS_CORE.set(address_index, address_store);
            addr_t := address_index;
          else
            if address_used = MAX_BITS then
              report "A Memory overflow has occurred.  Increase Size of A_BITS to increase memory size" severity failure;
            end if;
            ADDRESS_CORE.set(conv_integer(address_used), address_store);
            addr_t := conv_integer(address_used);
            address_used <= address_used + 1;
          end if;
          MEMORY_CORE.set(addr_t, Di(1), bw_i(1));
        else
          addr_t := conv_integer(wA_D(4));
          MEMORY_CORE.set(addr_t, Di(1), bw_i(1));
        end if;
      end if;
    end procedure Write_Array;

-- purpose: Main flow
  begin  -- process Main
    if RL=20 then
      if DLL='1' then
        Delay   <=3;
        Latency <=4;
      else
        Delay   <=1;
        Latency <=2;
      end if;
    elsif RL=25 then
      if DLL='1' then
        Delay   <=4;
        Latency <=5;
      else
        Delay   <=1;
        Latency <=2;
      end if;
    end if;

    if (posedge(K_i)) then
      shift;
      bw_i(0) <= not cBW;
      Di(0)   <= DQ;
      State;
      Write_Array("RISE");
      Read_Array;
    end if;

    if (negedge(K_i)) then
      shift;
      bw_i(0) <= not cBW;
      Di(0)   <= DQ;
      Write_Array("FALL");
      Read_Array;
    end if;
  end process Main;

-- purpose: Determine if TCK is running
-- type   : combinational
-- inputs : TCK
-- outputs: JTAG_Flag
  JTAG : process (TCK)
    variable TCK_Count : integer := 0;
  begin  
    if (POSEDGE(TCK)) then
      if (TCK_Count>=2) then
        JTAG_Flag <= 1;
      end if;
      if (TCK_Count<2) then
        TCK_Count := TCK_Count + 1;
      end if;
    end if;
  end process JTAG;

  C_int : process (K, nK)
  begin  
    if (K = '1') then
      K_i  <= '1';
    elsif nK = '1' then
      K_i  <= '0';
    end if;
  end process C_int;

-- purpose: Find time since last K clock edge
-- type   : combinational
-- inputs : K
-- outputs: K_time
  time_K_clk : process (K)
    variable K_Count   : integer := 0;
  begin
    if K_count < 4  then
      K_count := K_count + 1;
      K_time <= 1.0 ns;
    else
      K_time <= K'delayed'last_event;
    end if;
  end process time_K_clk;

-- purpose: Find time since last nK clock edge
-- type   : combinational
-- inputs : nK
-- outputs: nK_time
  time_nK_clk : process (nK)
    variable nK_Count   : integer := 0;
  begin
    if nK_count < 4  then
      nK_count := nK_count + 1;
      nK_time <= 1.0 ns;
    else
      nK_time <= nK'delayed'last_event;
    end if;
  end process time_nK_clk;
  
  Q_Clock : process (CQ_o, nCQ_o)
    begin
      if posedge(CQ_o) or posedge(nCQ_o) then
        for i in 7 downto 1 loop
          read_pos(i) <= read_pos(i-1);
        end loop;  -- i
        read_pos(0) <= re;
        if (read_pos(Delay) = '1') then
          Q_switch <= transport '1' after (tCQHQV);
          Q_switch <= transport '0' after (K_time + tCQHQX);
        else
          Q_switch <= transport '0' after (K_time + tCQHQX);
        end if;
      end if;
      Qi_o <= Qi;
    end process Q_Clock;

  CQ_Clock : process (K, nK)
    begin
      if JTAG_Flag = 0 then
        if (posedge(K)) then
          CQ_o  <= transport '1'after CHCQon;
          nCQ_o <= transport '0'after CHCQoff;
        end if;
        if (posedge(nK)) then
          nCQ_o <= transport '1'after CHCQon;
          CQ_o  <= transport '0'after CHCQoff;
        end if;
      else
        CQ_o  <= 'Z';
        nCQ_o <= 'Z';
      end if;
    end process CQ_Clock;

  CHQV <= tCHQV when DLL = '1' else tCHQV_off;
  CHQX <= tCHQX when DLL = '1' else tCHQX_off;
  CHCQon  <= 4*K_time + (CHQV - tCQHQV);
  CHCQoff <= 4*K_time + (CHQV - tCQHQV);
--  CHCQoff <= 4*K_time + (CHQX - tCQHQX);
  DLL  <= nDoff;
  CQ   <= CQ_o;
  nCQ  <= nCQ_o;
  QVLD <= read_pos(Delay) after tCQHQV;
  DQ   <= Qi_o when Q_switch = '1' else (others=>'Z');

END PLUS_CIO_B2;
